@@ -1,6 +1,5 @@
 /* Output dbx-format symbol table information from GNU compiler.
    Copyright (C) 1987, 1988 Free Software Foundation, Inc.
-   Hacked by Michael Tiemann (tiemann@mcc.com)
 
 This file is part of GNU CC.
 
@@ -206,6 +205,17 @@ dbxout_init (asm_file, input_file_name)
   dbxout_types (get_permanent_types ());
 }
 
+/* Change by Bryan Boreham, Kewill, Sun Aug 13 15:31:25 1989.
+   Added to support unexecing of compiler.    */
+
+void
+re_init_dbxout_for_unexec (asm_file, input_file_name)
+     FILE *asm_file;
+     char *input_file_name;
+{
+  asmfile = asm_file;
+}
+
 /* Continue a symbol-description that gets too big.
    End one symbol table entry with a double-backslash
    and start a new one, eventually producing something like
@@ -240,7 +250,8 @@ dbxout_type (type, full)
      tree type;
      int full;
 {
-  register tree fields, tem;
+  register tree fields, tem, method_vec;
+  tree *methods, *end;
   char *vfield_name = 0;
   tree virtual_basetype = 0;
 
@@ -407,7 +418,7 @@ dbxout_type (type, full)
 		fprintf (asmfile, "%d,",
 			 BITS_PER_UNIT * DECL_OFFSET (TYPE_NAME (basetype)));
 		CHARS (15);
-		dbxout_type (basetype);
+		dbxout_type (basetype, full);
 		putc (';', asmfile);
 	      }
 	    else
@@ -419,7 +430,7 @@ dbxout_type (type, full)
 
 		fprintf (asmfile, "%s:", IDENTIFIER_POINTER (name));
 		CHARS (2 + IDENTIFIER_LENGTH (name));
-		dbxout_type (basetype);
+		dbxout_type (basetype, full);
 		fprintf (asmfile, ",%d,%d;",
 			 BITS_PER_UNIT * DECL_OFFSET (TYPE_NAME (basetype)),
 			 TREE_INT_CST_LOW (TYPE_SIZE (basetype)) * DECL_SIZE_UNIT (basetype));
@@ -437,8 +448,7 @@ dbxout_type (type, full)
 	if (TREE_CODE (tem) == TYPE_DECL)
 	  continue;
 	/* Omit here the nameless fields that are used to skip bits.  */
-	else if (DECL_NAME (tem) != 0
-	    && (use_gdb_dbx_extensions || TREE_CODE (tem) != CONST_DECL))
+	else if (DECL_NAME (tem) != 0 && TREE_CODE (tem) != CONST_DECL)
 	  {
 	    /* Continue the line if necessary,
 	       but not before the first field.  */
@@ -458,16 +468,21 @@ dbxout_type (type, full)
 	      }
 
 	    dbxout_type (TREE_TYPE (tem), 0);
-	    if (TREE_CODE (tem) == VAR_DECL)
+	    if (TREE_CODE (tem) == VAR_DECL && TREE_STATIC (tem))
 	      {
 		if (use_gdb_dbx_extensions)
 		  {
-		    char *name = XSTR (XEXP (DECL_RTL (tem), 0), 0);
-		    if (name[0] == '*')
-		      name += 1;
+		    char *name = DECL_ASSEMBLER_NAME (tem);
+
 		    /* Adding 1 here only works on systems
-		       which flush an initial underscore.  */
-		    fprintf (asmfile, ":%s;", name+1);
+		       which flush an initial underscore from
+		       the .stabs entry.  This loses for static names
+		       which have an initial leading '_' on systems which
+		       don't use leading underscores.  */
+		    if (name[0] == '_')
+		      name += 1;
+
+		    fprintf (asmfile, ":%s;", name);
 		    CHARS (strlen (name));
 		  }
 		else
@@ -475,11 +490,13 @@ dbxout_type (type, full)
 		    fprintf (asmfile, ",0,0;");
 		  }
 	      }
-	    else if (TREE_CODE (tem) == CONST_DECL)
+#if 0
+	    else if (TREE_CODE (tem) == VAR_DECL || TREE_CODE (tem) == CONST_DECL)
 	      {
-		/* Only get here if using GDB's extensions to DBX format.  */
-		fprintf (asmfile, "~i%d;", TREE_INT_CST_LOW (DECL_INITIAL (tem)));
+		/* GDB 3.2 can't understand these declarations yet.  */
+		continue;
 	      }
+#endif
 	    else
 	      {
 		fprintf (asmfile, ",%d,%d;", DECL_OFFSET (tem),
@@ -489,24 +506,35 @@ dbxout_type (type, full)
 	  }
       /* C++: put out the method names and their parameter lists */
       /* We do constructors, destructor, if any, followed by the method names.  */
-      fields = use_gdb_dbx_extensions ? CLASSTYPE_FN_FIELDS (type) : NULL_TREE;
-      if (TREE_CODE (type) == RECORD_TYPE
-	  && (TYPE_HAS_DESTRUCTOR (type) | TYPE_HAS_CONSTRUCTOR (type))
-	  && use_gdb_dbx_extensions)
+      method_vec = use_gdb_dbx_extensions ? CLASSTYPE_METHOD_VEC (type) : NULL_TREE;
+      methods = 0;
+      end = 0;
+      if (use_gdb_dbx_extensions
+	  && TREE_CODE (type) == RECORD_TYPE
+	  && (TYPE_HAS_DESTRUCTOR (type) | TYPE_HAS_CONSTRUCTOR (type)))
 	{
 	  tree dtor;
 
+	  end = TREE_VEC_END (method_vec);
 	  /* Destructors lie in a special place.  */
 	  if (TYPE_HAS_DESTRUCTOR (type))
 	    {
-	      dtor = TREE_VALUE (fields);
+	      methods = &TREE_VEC_ELT (method_vec, 0);
+	      dtor = TREE_VEC_ELT (method_vec, 0);
 	      tem = TREE_CHAIN (dtor);
 	    }
+	  else if (TREE_VEC_LENGTH (method_vec) > 1)
+	    {
+	      methods = &TREE_VEC_ELT (method_vec, 1);
+	      dtor = NULL_TREE;
+	      tem = *methods;
+	    }
 	  else
 	    {
-	      dtor = NULL_TREE;
-	      tem = TREE_VALUE (fields);
+	      methods = end;
+	      tem = 0;
 	    }
+
 	  CHARS (2);
 
 	  if (tem)
@@ -529,15 +557,21 @@ dbxout_type (type, full)
 		     well as the name of the field before overloading, along
 		     with its parameter list.  */
 		  tree t;
+		  char c;
 
 		  CONTIN;
 		  /* Get to the FUNCTION_DECL */
 		  t = tem;
 		  dbxout_type (TREE_TYPE (t), 0); /* METHOD_TYPE */
+		  if (DECL_VIRTUAL_P (tem))
+		    c = '*';
+		  else if (DECL_STATIC_FUNCTION_P (tem))
+		    c = '?';
+		  else
+		    c = '.';
 		  fprintf (asmfile, ":%s;%c%c",
 			   IDENTIFIER_POINTER (DECL_NAME (t)),
-			   TREE_PRIVATE (tem) ? '0' : TREE_PROTECTED (tem) ? '1' : '2',
-			   DECL_VIRTUAL_P (tem) ? '*' : '.');
+			   TREE_PRIVATE (tem) ? '0' : TREE_PROTECTED (tem) ? '1' : '2', c);
 		  CHARS (IDENTIFIER_LENGTH (DECL_NAME (t)) + 5);
 		  if (DECL_VIRTUAL_P (tem))
 		    {
@@ -553,11 +587,18 @@ dbxout_type (type, full)
 		}
 	      putc (';', asmfile);
 	    }
-	  fields = TREE_CHAIN (fields);
+	  if (methods != end)
+	    methods++;
 	}
-      for (; fields; fields = TREE_CHAIN (fields))
+      else if (method_vec != NULL_TREE)
 	{
-	  tem = TREE_VALUE (fields);
+	  methods = &TREE_VEC_ELT (method_vec, 1);
+	  end = TREE_VEC_END (method_vec);
+	}
+
+      for (; methods != end; methods++)
+	{
+	  tem = *methods;
 
 	  if (tem)
 	    {
@@ -570,8 +611,8 @@ dbxout_type (type, full)
 	      else
 		{
 		  fprintf (asmfile, "%s::",
-			   IDENTIFIER_POINTER (TREE_PURPOSE (fields)));
-		  CHARS (IDENTIFIER_LENGTH (TREE_PURPOSE (fields)) + 3);
+			   IDENTIFIER_POINTER (DECL_ORIGINAL_NAME (tem)));
+		  CHARS (IDENTIFIER_LENGTH (DECL_ORIGINAL_NAME (tem)) + 3);
 		}
 
 	      for (; tem; tem = TREE_CHAIN (tem))
@@ -581,16 +622,22 @@ dbxout_type (type, full)
 		{
 		  /* @@ */
 		  tree t;
+		  char c;
 
 		  CONTIN;
 
 		  /* Get to the FUNCTION_DECL */
 		  t = tem;
 		  dbxout_type (TREE_TYPE (t), 0); /* METHOD_TYPE */
+		  if (DECL_VIRTUAL_P (tem))
+		    c = '*';
+		  else if (DECL_STATIC_FUNCTION_P (tem))
+		    c = '?';
+		  else
+		    c = '.';
 		  fprintf (asmfile, ":%s;%c%c",
 			   IDENTIFIER_POINTER (DECL_NAME (t)),
-			   TREE_PRIVATE (tem) ? '0' : TREE_PROTECTED (tem) ? '1' : '2',
-			   DECL_VIRTUAL_P (tem) ? '*' : '.');
+			   TREE_PRIVATE (tem) ? '0' : TREE_PROTECTED (tem) ? '1' : '2', c);
 		  CHARS (IDENTIFIER_LENGTH (DECL_NAME (t)) + 6);
 		  if (DECL_VIRTUAL_P (tem))
 		    {
@@ -617,7 +664,7 @@ dbxout_type (type, full)
 	  else if (TYPE_HAS_CONSTRUCTOR (type))
 	    putc ('+', asmfile);
 
-	  if (TYPE_VIRTUAL_P (type))
+	  if (CLASSTYPE_VSIZE (type))
 	    {
 	      tree t, v = DECL_NAME (CLASSTYPE_VFIELD (type));
 	      t = type;
@@ -774,8 +821,9 @@ dbxout_symbol (decl, local)
 
   if (local == 0)
     {
-      dbxout_tags (gettags ());
+      /* Send out the types first.  */
       dbxout_types (get_permanent_types ());
+      dbxout_tags (gettags ());
     }
 
   current_sym_code = 0;
@@ -832,8 +880,10 @@ dbxout_symbol (decl, local)
 	  && TYPE_NAME (TREE_TYPE (decl)) == decl)
 	return;
 #endif
-      /* Don't output the same typedef twice.  */
-      if (TREE_ASM_WRITTEN (decl))
+      /* Don't output the same typedef twice.
+         And don't output what language-specific stuff doesn't want output.  */
+      if (TREE_ASM_WRITTEN (decl)
+	  || lang_output_debug_info (TREE_TYPE (decl)) == 0)
 	return;
 
       /* Output typedef name.  */
@@ -855,6 +905,8 @@ dbxout_symbol (decl, local)
 	 and are output by dbxout_reg_parms and dbxout_parms.  */
       abort ();
 
+    case RESULT_DECL:
+      /* Named return value, treat like a VAR_DECL.  */
     case VAR_DECL:
       if (DECL_RTL (decl) == 0)
 	return;
@@ -903,7 +955,10 @@ dbxout_symbol (decl, local)
 	 S for static storage and file scope,
 	 V for static storage and local scope,
 	    for those two, use N_LCSYM if data is in bss segment,
-	    N_STSYM otherwise.  (N_FUN confuses GDB.)
+	    N_STSYM if in data segment, N_FUN otherwise.
+	    (We used N_FUN originally, then changed to N_STSYM
+	    to please GDB.  However, it seems that confused ld.
+	    Now GDB has been fixed to like N_FUN, says Kingdon.)
 	 no letter at all, and N_LSYM, for auto variable,
 	 r and N_RSYM for register variable.  */
 
@@ -923,15 +978,10 @@ dbxout_symbol (decl, local)
 
 	      if (!DECL_INITIAL (decl))
 		current_sym_code = N_LCSYM;
-#if 0  /* Note: N_FUN confuses GDB, since GDB expects it to start a new
-	  nest of N_LBRAC/N_RBRAC, etc.  But N_STSYM probably does not
-	  work either, since it relocates as data segment.
-	  Probably no standard N_ code works, so we must invent one.  */
 	      else if (TREE_READONLY (decl) && ! TREE_VOLATILE (decl))
 		/* This is not quite right, but it's the closest
 		   of all the codes that Unix defines.  */
 		current_sym_code = N_FUN;
-#endif
 	      else
 		current_sym_code = N_STSYM;
 	    }
@@ -968,7 +1018,11 @@ dbxout_symbol (decl, local)
 	      current_sym_value = INTVAL (XEXP (XEXP (XEXP (DECL_RTL (decl), 0), 0), 1));
 	    }
 
-	  type = build_pointer_type (TREE_TYPE (decl));
+	  /* Effectively do build_pointer_type, but don't cache this type,
+	     since it might be temporary whereas the type it points to
+	     might have been saved for inlining.  */
+	  type = make_node (POINTER_TYPE);
+	  TREE_TYPE (type) = TREE_TYPE (decl);
 	}
       else if (GET_CODE (DECL_RTL (decl)) == MEM
 	       && GET_CODE (XEXP (DECL_RTL (decl), 0)) == REG)
@@ -996,22 +1050,20 @@ dbxout_symbol (decl, local)
       /* One slight hitch: if this is a VAR_DECL which is a static
 	 class member, we must put out the mangled name instead of the
 	 DECL_NAME.  */
+      /* Note also that static member (variable) names DO NOT begin
+	 with underscores in .stabs directives.  */
       if (DECL_LANG_SPECIFIC (decl))
 	{
-	  name = XSTR (XEXP (DECL_RTL (decl), 0), 0);
-	  if (name[0] == '*')
+	  name = DECL_ASSEMBLER_NAME (decl);
+
+	  /* Adding 1 here only works on systems
+	     which flush an initial underscore.  */
+	  if (name[0] == '_')
 	    name += 1;
 	}
-      else
-	{
-	  name = IDENTIFIER_POINTER (DECL_NAME (decl));
-	}
+      else name = IDENTIFIER_POINTER (DECL_NAME (decl));
 
-      /* Adding 1 here only works on systems
-	 which flush an initial underscore.  */
-      fprintf (asmfile, ".stabs \"%s:",
-	       name + (DECL_CONTEXT (decl) != 0
-		       && TREE_CODE (DECL_CONTEXT (decl)) == RECORD_TYPE));
+      fprintf (asmfile, ".stabs \"%s:", name);
       if (letter) putc (letter, asmfile);
       dbxout_type (type, 0);
       dbxout_finish_symbol ();
@@ -1127,15 +1179,12 @@ dbxout_parms (parms)
 	  if (DECL_NAME (parms))
 	    {
 	      current_sym_nchars = 2 + IDENTIFIER_LENGTH (DECL_NAME (parms));
-
-	      FORCE_TEXT;
 	      fprintf (asmfile, ".stabs \"%s:P",
 		       IDENTIFIER_POINTER (DECL_NAME (parms)));
 	    }
 	  else
 	    {
 	      current_sym_nchars = 8;
-	      FORCE_TEXT;
 	      fprintf (asmfile, ".stabs \"(anon):P");
 	    }
 
@@ -1155,18 +1204,19 @@ dbxout_parms (parms)
 	  if (DECL_NAME (parms))
 	    {
 	      current_sym_nchars = 2 + IDENTIFIER_LENGTH (DECL_NAME (parms));
-
-	      FORCE_TEXT;
 	      fprintf (asmfile, ".stabs \"%s:p",
 		       IDENTIFIER_POINTER (DECL_NAME (parms)));
 	    }
 	  else
 	    {
 	      current_sym_nchars = 8;
-	      FORCE_TEXT;
 	      fprintf (asmfile, ".stabs \"(anon):p");
 	    }
 
+#if 0 /* This is actually the case in which a parameter
+	 is passed in registers but lives on the stack in a local slot.
+	 The address we are using is already correct, so don't change it.  */
+
 	  /* This is the case where the parm is passed as an int or double
 	     and it is converted to a char, short or float and stored back
 	     in the parmlist.  In this case, describe the parm
@@ -1178,6 +1228,7 @@ dbxout_parms (parms)
 	    current_sym_value += (GET_MODE_SIZE (TYPE_MODE (DECL_ARG_TYPE (parms)))
 				  - GET_MODE_SIZE (GET_MODE (DECL_RTL (parms))));
 #endif
+#endif /* 0 */
 
 	  dbxout_type (TREE_TYPE (parms), 0);
 	  dbxout_finish_symbol ();
@@ -1209,14 +1260,12 @@ dbxout_reg_parms (parms)
 	  if (DECL_NAME (parms))
 	    {
 	      current_sym_nchars = 2 + IDENTIFIER_LENGTH (DECL_NAME (parms));
-	      FORCE_TEXT;
 	      fprintf (asmfile, ".stabs \"%s:r",
 		       IDENTIFIER_POINTER (DECL_NAME (parms)));
 	    }
 	  else
 	    {
 	      current_sym_nchars = 8;
-	      FORCE_TEXT;
 	      fprintf (asmfile, ".stabs \"(anon):r");
 	    }
 	  dbxout_type (TREE_TYPE (parms), 0);
@@ -1245,14 +1294,12 @@ dbxout_reg_parms (parms)
 	      if (DECL_NAME (parms))
 		{
 		  current_sym_nchars = 2 + IDENTIFIER_LENGTH (DECL_NAME (parms));
-		  FORCE_TEXT;
 		  fprintf (asmfile, ".stabs \"%s:",
 			   IDENTIFIER_POINTER (DECL_NAME (parms)));
 		}
 	      else
 		{
 		  current_sym_nchars = 8;
-		  FORCE_TEXT;
 		  fprintf (asmfile, ".stabs \"(anon):");
 		}
 	      dbxout_type (TREE_TYPE (parms), 0);
@@ -1312,7 +1359,8 @@ dbxout_tags (tags)
       register tree type = TYPE_MAIN_VARIANT (TREE_VALUE (link));
       if (TREE_PURPOSE (link) != 0
 	  && ! TREE_ASM_WRITTEN (link)
-	  && TYPE_SIZE (type) != 0)
+	  && TYPE_SIZE (type) != 0
+	  && lang_output_debug_info (type))
 	{
 	  TREE_ASM_WRITTEN (link) = 1;
 	  current_sym_code = N_LSYM;
@@ -1325,6 +1373,20 @@ dbxout_tags (tags)
 		   ANON_AGGRNAME_P (TREE_PURPOSE (link)) ? "" : IDENTIFIER_POINTER (TREE_PURPOSE (link)));
 	  dbxout_type (type, 1);
 	  dbxout_finish_symbol ();
+
+/* Change by Bryan Boreham, Kewill, Fri Sep 22 16:57:42 1989.
+   Added to make sure all fully-output structs have typedefs.    */
+
+	  if (!ANON_AGGRNAME_P (TREE_PURPOSE (link)))
+	    {
+	      fprintf (asmfile, ".stabs \"%s:t",
+		       IDENTIFIER_POINTER (TREE_PURPOSE (link)));
+	      
+	      current_sym_code = N_LSYM;
+
+	      dbxout_type (type, 1);
+	      dbxout_finish_symbol ();
+	    }
 	}
     }
 }
@@ -1366,6 +1428,9 @@ dbxout_block (stmt, depth, args)
 	  break;
 
 	case LET_STMT:
+	  /* Ignore LET_STMTs for blocks never really used to make RTL.  */
+	  if (! TREE_USED (stmt))
+	    break;
 	  /* In dbx format, the syms of a block come before the N_LBRAC.  */
 	  dbxout_tags (STMT_TYPE_TAGS (stmt));
 	  dbxout_syms (STMT_VARS (stmt));
@@ -1378,16 +1443,26 @@ dbxout_block (stmt, depth, args)
 	     that final will define around the code in this block.  */
 	  if (depth > 0)
 	    {
+	      char buf[20];
 	      blocknum = next_block_number++;
-	      fprintf (asmfile, ".stabn %d,0,0,LBB%d\n", N_LBRAC, blocknum);
+	      ASM_GENERATE_INTERNAL_LABEL (buf, "LBB", blocknum);
+	      fprintf (asmfile, ".stabn %d,0,0,", N_LBRAC);
+	      assemble_name (asmfile, buf);
+	      fprintf (asmfile, "\n");
 	    }
 
-	  /* Output the interior of the block.  */
-	  dbxout_block (STMT_BODY (stmt), depth + 1, 0);
+	  /* Output the subblocks.  */
+	  dbxout_block (STMT_SUBBLOCKS (stmt), depth + 1, 0);
 
 	  /* Refer to the marker for the end of the block.  */
 	  if (depth > 0)
-	    fprintf (asmfile, ".stabn %d,0,0,LBE%d\n", N_RBRAC, blocknum);
+	    {
+	      char buf[20];
+	      ASM_GENERATE_INTERNAL_LABEL (buf, "LBE", blocknum);
+	      fprintf (asmfile, ".stabn %d,0,0,", N_RBRAC);
+	      assemble_name (asmfile, buf);
+	      fprintf (asmfile, "\n");
+	    }
 	}
       stmt = TREE_CHAIN (stmt);
     }
@@ -1403,14 +1478,31 @@ void
 dbxout_function (decl)
      tree decl;
 {
+  extern tree value_identifier;
+
   dbxout_symbol (decl, 0);
   dbxout_parms (DECL_ARGUMENTS (decl));
+  if (DECL_NAME (DECL_RESULT (decl)) != value_identifier)
+    dbxout_symbol (DECL_RESULT (decl), 1);
   dbxout_block (DECL_INITIAL (decl), 0, DECL_ARGUMENTS (decl));
   
   /* If we made any temporary types in this fn that weren't
      output, output them now.  */
   dbxout_types (get_temporary_types ());
 }
+
+/* GNU C++ extensions.  */
+
+/* At the start of the file, emit symbolic information to orient
+   GDB for this particular file's exception handling implementation.
+   EH_TYPE is the type name of the exception type.
+   EH_DECL is the global root of the exception handler stack.  */
+
+void
+dbxout_eh_init (eh_type, eh_decl)
+     tree eh_type, eh_decl;
+{
+}
 
 #else /* not DBX_DEBUGGING_INFO */
 
